# File name: service.py
#
# The service module defines a single class (Py3oWindowsService) that contains
# the functionality for running a py3o.renderserver as a Windows Service.
#
# To use this class, users must do the following:
# 1. Download and install the win32all package
#    (http://starship.python.net/crew/mhammond/win32/)
#
# 2. Open a command prompt and navigate to the directory where this file
#    is located.  Use one of the following commands to
#    install/start/stop/remove the service:
#    > service.py install
#    > service.py start
#    > service.py stop
#    > service.py remove
#    Additionally, typing "service.py" will present the user with all of the
#    available options for controlling an installed service
#
#    to configure the service use the commands provided by the service.py --help
#    this command is handled by our own program and not by the windows service framework
#
# Once installed, the service will be accessible through the Services
# management console just like any other Windows Service.  All service
# startup exceptions encountered by the Py3oWindowsService class will be
# viewable in the Windows event viewer (this is useful for debugging
# service startup errors); all application specific output or exceptions that
# are not captured by the standard logging mechanism should
# appear in the stdout/stderr logs.
#
# This module has been tested on Windows Server 2000, 2003, and Windows
# XP Professional.

from ConfigParser import SafeConfigParser
import optparse
import logging
import sys
import os
from os.path import *

import win32serviceutil
import win32service
import win32event
import win32process
import win32api
from win32com.client import constants
import _winreg


import pkg_resources
from pkg_resources import iter_entry_points
from pkg_resources import working_set, Environment

# the org name used to store your conf in the registry
organization = "py3o"
product_name = "py3o-renderserver"

def get_config():
    """find the config file path in the registry
    """
    class Config(object):
        pass

    config = Config()

    try:
        reg_key = _winreg.OpenKey(_winreg.HKEY_LOCAL_MACHINE,
                'SOFTWARE\\%s\\%s' % (organization, product_name))

        config.soffice_host = _winreg.QueryValueEx(reg_key, 'soffice_host')[0]
        config.soffice_port = _winreg.QueryValueEx(reg_key, 'soffice_port')[0]
        config.listen_port = _winreg.QueryValueEx(reg_key, 'listen_port')[0]
        config.listen_interface = _winreg.QueryValueEx(reg_key, 'listen_interface')[0]
        config.javalib = _winreg.QueryValueEx(reg_key, 'javalib')[0]
        config.driver = _winreg.QueryValueEx(reg_key, 'driver')[0]
        config.maxmem = _winreg.QueryValueEx(reg_key, 'maxmem')[0]
        #config.libdir = _winreg.QueryValueEx(reg_key, 'libdir')[0]

    except WindowsError, e:
        logging.exception(str(e))

    return config


def scan_directory(directory):
    distributions, errors = working_set.find_plugins(
            Environment([directory])
    )

    map(working_set.add, distributions)

    if len(errors) > 0:
        raise ValueError("Couldn't load %s" % errors)

class NullOutput(object):
    """a file-like object that behaves like a black hole.
    Does not consume memory and gives nothing back. Ever.
    """
    def noop(self, *args, **kw):
        pass

    write = writelines = close = seek = flush = truncate = noop

    def __iter__(self):
        return self

    def next(self):
        raise StopIteration

    def isatty(self):
        return False

    def tell(self):
        return 0

    def read(self, *args, **kw):
        return ''

    readline = read

    def readlines(self, *args, **kw):
        return list()



def set_config(options):
    """set the config file path in the registry
    """
    reg_key = _winreg.CreateKey(_winreg.HKEY_LOCAL_MACHINE,
            'SOFTWARE\\%s\\%s' % (organization, product_name))

    reg_val = _winreg.SetValueEx(
            reg_key, 'soffice_host', None, _winreg.REG_SZ, options.soffice_host)
    reg_val = _winreg.SetValueEx(
            reg_key, 'soffice_port', None, _winreg.REG_SZ, options.soffice_port)
    reg_val = _winreg.SetValueEx(
            reg_key, 'listen_port', None, _winreg.REG_SZ, options.listen_port)
    reg_val = _winreg.SetValueEx(
            reg_key, 'listen_interface', None, _winreg.REG_SZ, options.listen_interface)
    reg_val = _winreg.SetValueEx(
            reg_key, 'javalib', None, _winreg.REG_SZ, options.javalib)
    reg_val = _winreg.SetValueEx(
            reg_key, 'driver', None, _winreg.REG_SZ, options.driver)
    reg_val = _winreg.SetValueEx(
            reg_key, 'maxmem', None, _winreg.REG_SZ, options.maxmem)

    # reg_val = _winreg.SetValueEx(
    #         reg_key, 'libdir', None, _winreg.REG_SZ, options.libdir)

class Py3oWindowsService(win32serviceutil.ServiceFramework):
    """The Py3oWindowsService class contains all the functionality required
    for running a py3o renderserver as a Windows Service. The only
    user edits required for this class are located in the following class
    variables:

    _svc_name_:         The name of the service (used in the Windows registry).
                        DEFAULT: The capitalized name of the current directory.
    _svc_display_name_: The name that will appear in the Windows Service Manager.
                        DEFAULT: The capitalized name of the current directory.

    For information on installing the application, please refer to the
    documentation at the end of this module or navigate to the directory
    where this module is located and type "service.py" from the command
    prompt.
    """
    _svc_name_ = '%s' % (product_name)
    _svc_display_name_ = _svc_name_
    _svc_deps = list()

    def __init__(self, args):
        """set some usefull variables
        """

        # replace normal output channels by fake ones
        # to make sure we don't go over the buffer and inadvertently
        # kill our windows service
        #sys.stdout.close()
        #sys.stderr.close()
        sys.stdout = NullOutput()
        sys.stderr = NullOutput()

        win32serviceutil.ServiceFramework.__init__(self, args)
        #self.hWaitStop = win32event.CreateEvent(None, 0, 0, None)

    def SvcDoRun(self):
        """Called when the Windows Service runs."""

        self.ReportServiceStatus(win32service.SERVICE_START_PENDING)

        try:
            config = get_config()
            from py3o.renderserver.server import start_server
            self.ReportServiceStatus(win32service.SERVICE_RUNNING)
            start_server(config)

        except Exception, e:
            # something went wrong and we did not start correctly
            self.ReportServiceStatus(win32service.SERVICE_ERROR_CRITICAL)
            import servicemanager
            servicemanager.LogErrorMsg("The service could not start for the folloing reason: %s" % str(e))
            self.ReportServiceStatus(win32service.SERVICE_STOPPED)

    def SvcStop(self):
        """Called when Windows receives a service stop request."""
        self.ReportServiceStatus(win32service.SERVICE_STOP_PENDING)
        from twisted.internet import reactor
        reactor.stop()
        self.ReportServiceStatus(win32service.SERVICE_STOPPED)


def config():
    """write the configuration parameters of the py3o-renderserver service in the windows registry
    """
    optparser = optparse.OptionParser()

    optparser.add_option(
        "-a", "--sofficehost",
        dest="soffice_host",
        help="specify the open office hostname/ip address ADDR",
        metavar="ADDR",
        default="127.0.0.1")

    optparser.add_option(
        "-p", "--sofficeport",
        dest="soffice_port",
        help="specify the open office port PORT",
        metavar="PORT",
        default="8997")

    optparser.add_option(
        "-l", "--listenport",
        dest="listen_port",
        help="specify the PORT on which our service will listen",
        metavar="PORT",
        default=8994)

    optparser.add_option(
        "-i", "--listeninterface",
        dest="listen_interface",
        help="specify the INTERFACE on which our service will listen "
             "(default: all interfaces)",
        metavar="INTERFACE",
        default=None)

    optparser.add_option(
        "-d", "--driver",
        dest="driver",
        help="choose a driver between juno & pyuno",
        default='juno')

    optparser.add_option(
        "-j", "--java",
        dest="javalib",
        help="choose a jvm.dll/jvm.so to use if you are using the juno driver",
        default=None)

    optparser.add_option(
        "-m", "--maxmem",
        dest="maxmem",
        help="how much memory to give to the JVM if you are using juno driver, default is 150Mb",
        default='150')


    (options, args) = optparser.parse_args()
    set_config(options)


def setup():
    """basic win32 service setup: install or remove or update or start or stop the service
    """

    # The following are the most common command-line arguments that are used
    # with this module:
    #  service.py install (Installs the service with manual startup)
    #  service.py --startup auto install (Installs the service
    #  with auto startup)
    #  service.py start (Starts the service)
    #  service.py stop (Stops the service)
    #  service.py remove (Removes the service)
    #
    # For a full list of arguments, simply type "service.py".

    win32serviceutil.HandleCommandLine(Py3oWindowsService)



#!/usr/bin/env python
# -*- coding: iso-8859-1 -*-
# setup.py

from distutils.core import setup
from distutils.core import Command
import os, glob, sys
import py2exe

dist_dir = "dist\\py3o.renderserver-service"

# this is from py2exe but imported only in a sub-command and thus not
# automatically detected by py2exe
# we need this for registry reading
include_modules = ['_winreg', 'decimal']
include_packages = ["encodings", "twisted"]
options = dict()
	
# Some things we do not need in our build
dll_excludes = list()

excludes = [
		"pywin",
		"pywin.debugger",
		"pywin.debugger.dbgcon",
		"pywin.dialogs",
		"pywin.dialogs.list",
		"ltihooks",
		"uno",
		"pyuno",
		"unohelper",
		"com.sun.star.uno",
		"com.sun.star.connection",
		"com.sun.star.lang",
		"com.sun.star.beans",
        "simplejson",
        "dateutil",
        "jpype"
		]

def __removedirs(top):
	for root, dirs, files in os.walk(top, topdown=False):
		for name in files:
			os.remove(os.path.join(root, name))
		for name in dirs:
			os.rmdir(os.path.join(root, name))
	if os.path.exists(top):
		os.rmdir(top)

def __removefiles(top, suffix):
	for root, dirs, files in os.walk(top):
		for name in files:
			if name.endswith(suffix):
				os.remove(os.path.join(root, name))


if "clean" in sys.argv:
    top = os.getcwd()

    __removedirs(os.path.join(top, dist_dir))
    __removedirs(os.path.join(top, 'build'))
    __removefiles(top, 'pyc')

# If run without args, build win32 executables.
if len(sys.argv) == 1:
    sys.argv.append("py2exe")
    sys.argv.append("-q")

class Target:
    def __init__(self, **kw):
        self.__dict__.update(kw)
        # for the versioninfo resources
        #self.version = "0.5.0"
        self.company_name = ""
        self.copyright = "Florent AIDE 2010"
        self.name = "py3oservice"

class MyCommand(Command):
    depend_files = []

    def need_build(self, file):
        for dep_file in self.depend_files:
            if newer(dep_file, file):
                return True

class NSI(MyCommand):

    user_options = [('version', 'V', "Set version")]

    def initialize_options(self):
        self.version = VERSION

    def finalize_options(self):
        pass

    def need_py2exe(self):
        return True

    def run(self):
        call_args = [MAKENSIS_EXECUTABLE, '/DVersion=%s' % VERSION, NSI_FILE]
        Popen(args=call_args)

# an xml template file for winXP control
manifest_template = '''
<?xml version="1.0" encoding="UTF-8" standalone="yes"?>
<assembly xmlns="urn:schemas-microsoft-com:asm.v1" manifestVersion="1.0">
<assemblyIdentity
    version="5.0.0.0"
    processorArchitecture="x86"
    name="%(prog)s"
    type="win32"
/>
<description>%(prog)s Program</description>
<dependency>
    <dependentAssembly>
        <assemblyIdentity
            type="win32"
            name="Microsoft.Windows.Common-Controls"
            version="6.0.0.0"
            processorArchitecture="X86"
            publicKeyToken="6595b64144ccf1df"
            language="*"
        />
    </dependentAssembly>
</dependency>
</assembly>
'''

RT_MANIFEST = 24
#RT_ICON = 3

py3o_renderservice = Target(
	description = "py3o.renderserver windows service",
	modules = ['py3o.renderserver.service'],
	icon_resources = [],
	other_resources = [(RT_MANIFEST, 1, manifest_template % dict(prog="py3oservice"))],
	dest_base = "py3o_renderservice")

#add_data = [
#        ('.', (
#            'py3oserver.ini-sample',
#            )),
#        ]
add_data = []

setup(
	options = {"py2exe": {
						"compressed": 1,
						"optimize": 2,
						"packages": include_packages,
						"includes": include_modules,
						"excludes": excludes,
						"dll_excludes": dll_excludes,
						"dist_dir": dist_dir,
						},
			},

	zipfile = "libs\\shared.zip",
	console = [],
	windows = [],
	service = [py3o_renderservice],
	data_files = add_data,
    cmdclass = dict(nsi=NSI),
	)





import time
from ConfigParser import SafeConfigParser
import optparse
import logging

import sys
import os
from os.path import *

import win32serviceutil
import win32service
import win32event
import win32process
import win32api
from win32com.client import constants
import _winreg
from subprocess import Popen

# the org name used to store your conf in the registry
organization = "py3o"
product_name = "py3o-sofficeserver"


class NullOutput(object):
    """a file-like object that behaves like a black hole.
    Does not consume memory and gives nothing back. Ever.
    """
    def noop(self, *args, **kw):
        pass
    
    write = writelines = close = seek = flush = truncate = noop
    
    def __iter__(self):
        return self

    def next(self):
        raise StopIteration

    def isatty(self):
        return False
    
    def tell(self):
        return 0

    def read(self, *args, **kw):
        return ''

    readline = read
    
    def readlines(self, *args, **kw):
        return list()

def get_config():
    """find the config file path in the registry
    """
    config = dict()

    try:
        reg_key = _winreg.OpenKey(_winreg.HKEY_LOCAL_MACHINE,
                'SOFTWARE\\%s\\%s' % (organization, product_name))

        config["soffice_path"] = _winreg.QueryValueEx(reg_key, 'soffice_path')[0]
        config["soffice_host"] = _winreg.QueryValueEx(reg_key, 'soffice_host')[0]
        config["soffice_port"] = int(_winreg.QueryValueEx(reg_key, 'soffice_port')[0])

    except WindowsError, e:
        logging.exception(str(e))

    return config

def set_config(options):
    """set the config file path in the registry
    """
    reg_key = _winreg.CreateKey(_winreg.HKEY_LOCAL_MACHINE,
            'SOFTWARE\\%s\\%s' % (organization, product_name))

    reg_val = _winreg.SetValueEx(
            reg_key, 'soffice_path', None, _winreg.REG_SZ, options.soffice_path)
    reg_val = _winreg.SetValueEx(
            reg_key, 'soffice_host', None, _winreg.REG_SZ, options.soffice_host)
    reg_val = _winreg.SetValueEx(
            reg_key, 'soffice_port', None, _winreg.REG_SZ, options.soffice_port)

class Py3oSofficeService(win32serviceutil.ServiceFramework):
    """The Py3oSofficeService class contains all the functionality required
    for running a Open Office instance as a Windows Service.
    """
    _svc_name_ = '%s' % (product_name)
    _svc_display_name_ = _svc_name_
    _svc_deps = list()

    def __init__(self, args):
        """set some usefull variables
        """

        # replace normal output channels by fake ones
        # to make sure we don't go over the buffer and inadvertently
        # kill our windows service
        sys.stdout.close()
        sys.stderr.close()
        sys.stdout = NullOutput()
        sys.stderr = NullOutput()

        win32serviceutil.ServiceFramework.__init__(self, args)
        self.hWaitStop = win32event.CreateEvent(None, 0, 0, None)

    def SvcDoRun(self):
        """Called when the Windows Service runs."""

        self.ReportServiceStatus(win32service.SERVICE_START_PENDING)

        self.serv_process = self.runoo()
        pcs = win32process.EnumProcesses()

        if self.serv_process.pid in pcs:
            self.ReportServiceStatus(win32service.SERVICE_RUNNING)
            # here we should check regularly that our PID for OO has not vanished...
            win32event.WaitForSingleObject(self.hWaitStop, win32event.INFINITE)

        else:
            # something went wrong and we did not start correctly
            self.ReportServiceStatus(win32service.SERVICE_ERROR_CRITICAL)
            self.ReportServiceStatus(win32service.SERVICE_STOPPED)            

    def SvcStop(self):
        """Called when Windows receives a service stop request."""

        self.ReportServiceStatus(win32service.SERVICE_STOP_PENDING)

        try:
            win32api.TerminateProcess(int(self.serv_process._handle), -1)
        except:
            # TODO if we see errors one day, maybe do some logging... ?
            pass  

        win32event.SetEvent(self.hWaitStop)
        self.ReportServiceStatus(win32service.SERVICE_STOPPED)

    def runoo(self):
        """Bootsrap an Uno application environment by starting a soffice server
        and then connecting to it.
        """
        call_flags = 0
        call_flags |= win32process.CREATE_NO_WINDOW

        config = get_config()

        soffice_bin = config['soffice_path']
        host = config["soffice_host"]
        tcp_port = config["soffice_port"]

        cmdArray = ('%s' % soffice_bin,
                "-nologo",
                "-norecover",
                "-norestore",
                "-invisible",
                "-headless",
                "-nocrashreport",
                "-nofirststartwizard",
                "-nodefault",
                "-accept=socket,host=%s,port=%s;urp;" % (host, tcp_port))

        p = Popen(cmdArray, creationflags=call_flags)
        time.sleep(1.0)

        return p

def setup():
    """basic win32 service setup: install or remove or update or start or stop the service
    """
    win32serviceutil.HandleCommandLine(Py3oSofficeService)

def config():
    """write the configuration parameters of the OpenOffice service in the windows registry
    """
    optparser = optparse.OptionParser()

    optparser.add_option(
        "-s", "--sofficepath",
        dest="soffice_path",
        help="specify the open office file path FILE",
        metavar="FILE",
        default=None)

    optparser.add_option(
        "-a", "--sofficehost",
        dest="soffice_host",
        help="specify the open office hostname/ip address ADDR",
        metavar="ADDR",
        default="127.0.0.1")

    optparser.add_option(
        "-p", "--sofficeport",
        dest="soffice_port",
        help="specify the open office port PORT",
        metavar="PORT",
        default="8997")

    (options, args) = optparser.parse_args()

    set_config(options)
    sys.exit(0)





# -*- encoding: utf-8 -*-
from setuptools import setup, find_packages

version = '0.5.2'

setup(
    name='py3o.renderserver',
    version=version,
    description=("An easy solution to transform "
                 "libreoffice/openoffice documents to supported formats"),
    long_description=open("README.rst").read(),
    classifiers=[
        "Programming Language :: Python",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
    keywords='LibreOffice OpenOffice PDF',
    author='Florent Aide',
    author_email='florent.aide@gmail.com',
    url='http://bitbucket.org/faide/py3o.renderserver',
    license='BSD License',
    packages=find_packages(exclude=['ez_setup', 'examples', 'tests']),
    namespace_packages=['py3o'],
    include_package_data=True,
    zip_safe=True,
    install_requires=[
        'setuptools',
        'pyf.station >= 2.0.1',
        'pyjon.utils >= 0.6',
        'py3o.formats',
        'py3o.renderers.juno >= 0.8.1',
    ],
    entry_points=dict(
        # -*- Entry points: -*-
        console_scripts=[
            'start-py3o-renderserver = py3o.renderserver.server:cli_server',
            'config-ooservice = ooservice.ooservice:config',
            'setup-ooservice = ooservice.ooservice:setup',
            'config-py3oservice = py3o.renderserver.service:config',
            'setup-py3oservice = py3o.renderserver.service:setup',
        ],
    ),
    test_suite = 'nose.collector',
)



# -*- encoding: utf-8 -*-
import logging

from pyf.transport.packets import Packet
from pyf.station import FlowServer
from pyf.station.utils import base64encoder
from pyf.station.utils import base64decoder
from pyf.station.utils import file_to_packets
from pyf.station.utils import packets_to_file

from pyjon.utils import get_secure_filename

import json
import os
import sys
import optparse
from py3o.formats import Formats
from py3o.renderers.juno import Convertor
from twisted.internet.threads import deferToThread

logging.basicConfig(level=logging.DEBUG)
formats = Formats()
all_formats = formats.get_known_format_names()


class DataHandler(object):
    # those will be set in start server and be shared globally for the
    # whole process
    soffice_port = None
    soffice_host = None

    def __init__(self, flow, client):
        self.flow = flow
        self.client = client

    def render(self, header, flow, callback):
        source_format = header.get('source_format', 'odt')
        pdf_options = None
        pdf_options_str = header.get('pdf_options')
        if pdf_options_str:
            pdf_options = json.loads(pdf_options_str)
        input_filename = get_secure_filename(
            suffix=".%s" % source_format.lower()
        )
        output_filename = get_secure_filename()

        try:
            # warning usage of the callback from inside this thread will bork!!
            packets_to_file(flow, input_filename, callback)
            format_ = formats.get_format(header.target_format.lower())

            convertor = Convertor(self.soffice_host, self.soffice_port)
            logging.info(
                'Connecting to libreoffice %s:%s to convert document to %s',
                self.soffice_host, self.soffice_port, format_.odfname)
            if pdf_options:
                logging.info('PDF export options: %s', pdf_options)
            convertor.convert(
                input_filename,
                output_filename,
                format_.odfname,
                pdf_options)

        finally:
            os.unlink(input_filename)

        return output_filename

    def rendered(self, output_filename, callback):
        callback("render_status", "ok", filename=output_filename)

        for data in base64encoder(file_to_packets(open(output_filename, 'rb'))):
            self.client.message(data)

        self.client.success("Process ended successfully")
        logging.info('Process ended successfully')
        os.unlink(output_filename)

    def renderfailed(self, err):
        self.client.error(err)
        logging.error('Error received from libreoffice: %s', err)

    def handle_data(self):
        header = self.flow.next()

        if header.action == 'render':
            callback = lambda key, value, **kwargs: self.client.message(
                Packet(dict(type='appinfo', key=key, value=value, **kwargs))
            )

            if header.target_format.lower() in all_formats:
                d = deferToThread(
                    self.render, header, base64decoder(self.flow), callback
                )
                d.addCallback(self.rendered, callback)
                d.addErrback(self.renderfailed)

            else:
                callback("render_status", "nok", client=self.client)
                raise ValueError("Unsupported format %s" % header.target_format)

        else:
            callback("render_status", "nok", client=self.client)
            raise ValueError("Unsupported action %s" % header.action)

        logging.info("end of flow...")


def cli_server():
    optparser = optparse.OptionParser()

    optparser.add_option(
        "-a", "--sofficehost",
        dest="soffice_host",
        help="specify the open office hostname/ip address ADDR",
        metavar="ADDR",
        default="127.0.0.1")

    optparser.add_option(
        "-p", "--sofficeport",
        dest="soffice_port",
        help="specify the open office port PORT",
        metavar="PORT",
        default="8997")

    optparser.add_option(
        "-l", "--listenport",
        dest="listen_port",
        help="specify the PORT on which our service will listen",
        metavar="PORT",
        default=8994)

    optparser.add_option(
        "-i", "--listeninterface",
        dest="listen_interface",
        help="specify the INTERFACE on which our service will listen "
             "(default: all interfaces)",
        metavar="INTERFACE",
        default=None)

    optparser.add_option(
        "-d", "--driver",
        dest="driver",
        help="choose a driver between juno & pyuno",
        default='juno')

    optparser.add_option(
        "-j", "--java",
        dest="javalib",
        help="choose a jvm.dll/jvm.so to use if you are using the juno driver",
        default=None)

    optparser.add_option(
        "-u", "--ure",
        dest="ure_base",
        help="choose a ure base dir, ex: /usr/lib on ubuntu/libreoffice",
        default=None)

    optparser.add_option(
        "-o", "--office",
        dest="office_base",
        help="choose a office base dir, "
        "ex: /usr/lib64/libreoffice on ubuntu64/libreoffice",
        default=None)

    optparser.add_option(
        "-m", "--maxmem",
        dest="maxmem",
        help="how much memory to give to the JVM if you are using juno driver"
        ", default is 150Mb",
        default=150)

    optparser.add_option(
        "-t", "--office-type",
        dest="office_type",
        help="Windows only option! By default we try to find LibreOffice, "
        "but if you need to use OpenOffice, "
        "set this option to --office-type=openoffice",
        default='libreoffice')

    (options, args) = optparser.parse_args()

    start_server(options)


def start_server(options):
    if os.name in ('nt', 'os2', 'ce'):
        # on windows we expect to find LibreOffice
        # or OpenOffice in the registry
        if options.office_type == 'openoffice':
            base_ooregkey = '''SOFTWARE\\OpenOffice.org\\OpenOffice.org'''

        elif options.office_type == 'libreoffice':
            base_ooregkey = (
                '''SOFTWARE\\Wow6432Node\\LibreOffice\\LibreOffice'''
            )

        else:
            raise ValueError(
                '%s is not a valid office type' % options.office_type)

        import _winreg
        # TODO: make sure that we do not explode in flight when the key is
        # not present...
        # TODO: maybe we could use something else than just LOCAL_MACHINE ?
        base_key = _winreg.HKEY_LOCAL_MACHINE
        try:
            key = _winreg.OpenKey(base_key, base_ooregkey)
            # first entry should be 2.3 or something similar
            version = _winreg.EnumKey(key, 0)
            # set this variable for later use
            office_version = version

            # _winreg.CloseKey(key)
            key = _winreg.OpenKey(
                base_key, "%s\\%s" % (base_ooregkey, version)
            )

        except WindowsError:
            # some error occured:
            # Open Office is not installed on this machine!
            raise ValueError(
                'OpenOffice is not installed on this Windows host')

        soffice_bin, key_type = _winreg.QueryValueEx(key, 'Path')
        soffice_dir = os.path.dirname(soffice_bin)
        soffice_rootdir = os.path.split(soffice_dir)[0]
        # TODO: we must test this on Windows... this is only a placeholder
        # for the moment
        ure_base = soffice_rootdir

        # add to search path for uno import!
        sys.path.insert(0, soffice_dir)
        if options.javalib is None and options.driver == 'juno':
            # we need java for juno support so we try to find it ourselves
            try:
                base_java_key = (
                    '''SOFTWARE\\JavaSoft\\Java Runtime Environment'''
                )
                key = _winreg.OpenKey(base_key, base_java_key)
                numversions = _winreg.QueryInfoKey(key)[0]
                last_version = _winreg.EnumKey(key, numversions - 1)
                if numversions > 1:
                    logging.warning(
                        'Multiple Java versions found, '
                        'using the highest: %s' % last_version)

                javakey = _winreg.OpenKey(
                    base_key, "%s\\%s" % (
                        base_java_key, last_version
                    )
                )
                jvm, key_type = _winreg.QueryValueEx(javakey, 'RuntimeLib')
            except WindowsError:
                # some error occured:
                # Java is not (correctly?) installed on this machine!
                raise ValueError('Java is not installed on this Windows host')

        else:
            jvm = options.javalib

    else:
        # Here we are on Linux, nothing can be autodetected so we need explicit
        # values from the command line parser
        jvm = options.javalib
        soffice_rootdir = options.office_base
        ure_base = options.ure_base

    if not jvm:
        raise ValueError(
            "You must specify a JVM library "
            "(run 'config-py3oservice')")

    jvm_path = os.path.abspath(jvm)

    if not os.path.exists(jvm_path):
        raise ValueError("The specified JVM library does not exist"
                         ", can not proceed")

    if options.driver == 'juno':
        from py3o.renderers.juno import start_jvm
        start_jvm(
            jvm,
            soffice_rootdir,
            ure_base,
            int(options.maxmem)
        )
    else:
        raise ValueError("Renderer not yet supported")

    DataHandler.soffice_host = options.soffice_host
    DataHandler.soffice_port = options.soffice_port
    factory = FlowServer(DataHandler)

    # we install the selectreactor because we have intermittent but reccurent
    # errors when using the epoll reactor which is the default on linux
    from twisted.internet import selectreactor
    selectreactor.install()

    from twisted.internet import reactor
    logging.info(
        "listening on %s:%s",
        options.listen_interface or '*', options.listen_port)
    if options.listen_interface:
        reactor.listenTCP(
            int(options.listen_port),
            factory,
            interface=options.listen_interface)
    else:
        reactor.listenTCP(int(options.listen_port), factory)
    reactor.run(installSignalHandlers=1)
